METADATA_STATE_KEYS = [
    # Projection
    "spherical",
    # Cropping bounds
    "longitude_bnds",
    "latitude_bnds",
    # Vertical scaling
    "vertical_scaling",
    "vertical_scaling_topo",
    # Grid Box
    "nb_grid_line_per_degree",
    "show_grid",
    # Earth visibility
    "show_earth_core",
    # Coast line regions
    "coast_active_regions",
    # Do we have topo
    "topo_ui",
    # Visibility of meshes
    "show_surface",
    "show_segment",
    "show_terrain",
    "show_rivers",
    # Shadow of meshes
    "light_surface",
    "light_segment",
    "light_terrain",
    "light_rivers",
    # Opacity of meshes
    "surface_opacity",
    "segment_opacity",
    "terrain_opacity",
    "rivers_opacity",
    # Coloring information
    "color_by",
    "color_min",
    "color_max",
    "color_preset",
    # Processing pipeline
    "subdivide",
    "nb_contours",
    "use_formula",
    "formula",
]
